#!/usr/bin/env python

"""
View annotated sources.

Copyright (C) 2006, 2007, 2010, 2011, 2012, 2013 Paul Boddie <paul@boddie.org.uk>

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
details.

You should have received a copy of the GNU General Public License along with
this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from micropython.common import *
from micropython.data import *
from micropython.errors import *
from os.path import exists, extsep, join
import compiler.ast
import sys
import os
import textwrap

try:
    set
except NameError:
    from sets import Set as set

# Classes.

# HTML-related output production.

html_header = """<?xml version="1.0" encoding="iso-8859-15"?>
<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Transitional//EN" "http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd">
<html xmlns="http://www.w3.org/1999/xhtml">
<head>
  <title>%(title)s</title>
  <style type="text/css">
    html {
      background-color: black; color: white;
    }

    body {
      padding-bottom: 4em;
      font-size: 14pt; font-family: monospace;
      background-color: black; color: white;
      margin-top: 2em;
    }

    a {
      text-decoration: none;
    }

    .nowrap { white-space: nowrap; }
    .label { font-size: smaller; }

    .class { margin-top: 1em; margin-bottom: 1em; }
    .function { margin-top: 1em; margin-bottom: 1em; }
    .body { padding-left: 2em; }
    .for, .if, .tryexcept, .tryfinally, .while { margin-bottom: 1em; }
    .keyword { color: yellow; }
    .comment { color: blue; }
    .class-name { color: cyan; }
    .function-name { color: cyan; }
    .specific-ref { color: #07F; }
    .str { color: #FF00FF; }
    .doc { color: #FF00FF; margin-top: 1em; margin-bottom: 1em; }
    .doc.module { font-size: smaller; }

    .popup {
      display: none;
      position: absolute;
      top: 3ex; left: 0;
      color: white;
      z-index: 3;
    }

    .attributes-popup,
    .types-popup {
      display: none;
      position: absolute;
      bottom: 3ex; left: 0;
      color: white;
      z-index: 3;
    }

    .no-targets {
      background-color: #d00;
      color: white;
    }

    .any-target {
      background-color: #d60;
      color: white;
    }

    .attr,
    .accessor,
    .name,
    .operation {
      position: relative;
      background-color: #300;
      color: white;
    }

    .attr:hover,
    .accessor:hover,
    .name:hover,
    .operation:hover {
      background-color: #500;
      padding-top: 0.5ex;
      padding-bottom: 0.5ex;
      z-index: 2;
    }

    .attr:hover .attributes-popup,
    .accessor:hover .types-popup,
    .name:hover .popup,
    .operation:hover .popup {
      display: block;
    }

    .attrnames,
    .opnames,
    .scope,
    .typenames {
      padding: 0.5em;
      background-color: #700;
    }

    .name a,
    .attrnames a,
    .opnames a,
    .scope a {
      color: white;
    }

    .summary-class {
      vertical-align: top;
    }

    th.summary-class {
      font-weight: normal;
    }

    .summary-attr {
      background-color: #070;
    }

    .summary-interface,
    .summary-attr {
      font-size: smaller;
    }

    .summary-interface.complete {
      background-color: #050;
    }

    .summary-attr-absent {
      border-left: 0.2em solid #070;
      font-size: small;
    }

    .summary-class-attr {
      background-color: #007;
      font-size: smaller;
    }

    .summary-class-attr-absent {
      border-left: 0.2em solid #007;
      font-size: small;
    }

    .summary-ref {
      color: white;
    }

  </style>
</head>
<body>
"""

html_footer = """</body>
</html>
"""

# Utility classes.

class Writer:

    "A utility class providing useful HTML output methods."

    # Methods which return strings.

    def _text(self, text):
        return text.replace("&", "&amp;").replace("<", "&lt;").replace(">", "&gt;")

    def _attr(self, attr):
        return self._text(attr).replace("'", "&apos;").replace('"', "&quot;")

    def _url(self, url):
        return self._attr(url).replace("#", "%23").replace("-", "%2d")

    # Methods which write to the stream.

    def _span_start(self, classes=None):
        self.stream.write("<span class='%s'>" % (classes or ""))

    def _span_end(self):
        self.stream.write("</span>")

    def _span(self, value, classes=None):
        self._span_start(classes)
        self.stream.write(self._text(value))
        self._span_end()

    def _name_start(self, classes=None):
        self._span_start(classes or "name")

    _name_end = _span_end

    def _name(self, name, classes=None):
        self._name_start(classes)
        self.stream.write(self._text(name))
        self._name_end()

    def _popup_start(self, classes=None):
        self._span_start(classes or "popup")

    _popup_end = _span_end

    def _comment(self, comment):
        self._span("# %s" % comment, "comment")
        self.stream.write("\n")

    def _reserved(self, token, classes, leading=0, trailing=1):
        if leading:
            self.stream.write(" ")
        self._span(token, classes)
        if trailing:
            self.stream.write(" ")

    def _keyword(self, kw, leading=0, trailing=1):
        self._reserved(kw, "keyword", leading, trailing)

    def _doc(self, node, classes=None):
        if node.doc is not None:
            self._docstring(node.doc, classes)

    def _docstring(self, s, classes=None):
        self.stream.write("<pre class='doc%s'>" % (classes and " %s" % classes or ""))
        self.stream.write('"""')
        output = textwrap.dedent(s.replace('"""', '\\"\\"\\"'))
        self.stream.write(self._text(output))
        self.stream.write('"""')
        self.stream.write("</pre>\n")

    def _object_name_def(self, module, obj, classes=None):

        """
        Link to the summary for 'module' using 'obj'. The optional 'classes'
        can be used to customise the CSS classes employed.
        """

        if isinstance(obj, Class) or (isinstance(obj, Function) and obj.is_method()):
            self._summary_link(module.full_name(), obj.full_name(), obj.name, classes)
        else:
            self._span(obj.name, classes)

    def _object_name_ref(self, module, obj, name=None, classes=None):

        """
        Link to the definition for 'module' using 'obj' with the optional 'name'
        used as the label (instead of the name of 'obj'). The optional 'classes'
        can be used to customise the CSS classes employed.
        """

        self._name_link(module.full_name(), obj.full_name(), name or obj.name, classes)

    def _summary_link(self, module_name, full_name, name, classes=None):
        self._name_link("%s-summary" % module_name, full_name, name, classes)

    def _name_link(self, module_name, full_name, name, classes=None):
        self.stream.write("<a class='%s' href='%s%sxhtml#%s'>%s</a>" % (
            classes or "specific-ref", module_name, os.path.extsep,
            self._attr(full_name), self._text(name)))

    def _module_link(self, module_name, label, classes=None):
        self.stream.write("<a class='%s' href='%s%sxhtml'>%s</a>" % (
            classes or "name", module_name, os.path.extsep,
            self._text(label)))

    def _scope(self, scope, attr):
        self.stream.write("<div class='scope'>"
            "<span class='label'>scope</span><br />%s<br/>" % scope)
        values = self._values_to_attribute_names(attr)
        if values:
            self.stream.write("<span class='label'>values</span><br />")
            self._attribute_list(values)
        self.stream.write("</div>\n")

    def _assname(self, name, node):
        if node.flags == "OP_DELETE":
            self._keyword("del")
        self._span_start("assname")
        if not self._attrcombined(name, node):
            self._span(name)
        self._span_end()

    def _op(self, symbol, name=None, leading=0, trailing=1):
        if leading:
            self.stream.write(" ")
        self._span_start(name and "operation" or None)
        self._span(symbol, "operator")
        if name is not None:
            self._popup_start()
            self.stream.write("<div class='opnames'>")
            self._name_link("operator", "operator.%s" % name, name)
            self.stream.write("</div>\n")
            self._popup_end()
        # NOTE: Handle "is" and "in".
        self._span_end()
        if trailing:
            self.stream.write(" ")

    def _names_list_start(self, label, classes):
        self.stream.write("<div class='%s'><span class='label'>%s</span><br />" % (classes, label))

    def _names_list_end(self):
        self.stream.write("</div>\n")

    def _names_list(self, names, label, classes):
        if not names:
            return
        names = list(names)
        names.sort()

        self._names_list_start(label, classes)
        first = True
        for name in names:
            if not first:
                self.stream.write("<br />")
            self.stream.write(self._text(name))
            first = False
        self._names_list_end()

    def _attrcombined(self, name, node):
        attrcombined = node._attrcombined and node._attrcombined.get(name) or []

        # Since assigned values will not be shown, produce a sorted list of
        # distinct attribute name lists.

        all_attrnames = set()
        for usage in attrcombined:
            if usage:
                all_attrnames.add(tuple(usage.keys()))

        if not all_attrnames:
            return False

        all_attrnames = list(all_attrnames)
        all_attrnames.sort()

        # Write the lists of attribute names.

        self._name_start()
        self.stream.write(name)
        self._popup_start()
        for attrnames in all_attrnames:
            self._attrnames(attrnames)
        self._popup_end()
        self._name_end()
        return True

    def _attrnames(self, attrnames):
        self._names_list(attrnames, "attributes", "attrnames")

    def _typenames(self, typenames):
        self._names_list(typenames, "types", "typenames")

    def _accessor_start(self, target_names):
        if target_names:
            self._span_start("accessor")
            self._popup_start("types-popup")
            self._typenames(target_names)
            self._popup_end()

    def _accessor_end(self, target_names):
        if target_names:
            self._span_end()

    def _values_to_attribute_names(self, attr):

        "Get the output form of the values referenced by 'attr'."

        if isinstance(attr, Const):
            return [(repr(attr.get_value()), attr)]
        elif isinstance(attr, Instance):
            return []

        have_instances = False
        values = []
        for v in attr.get_values():
            if isinstance(v, Const):
                values.append((repr(v.get_value()), v))
            elif not isinstance(v, Instance):
                values.append((v.full_name(), v))
            else:
                have_instances = True

        if have_instances:
            values.append(("...", None))

        values.sort()
        return values

    def _attribute_value_to_name(self, attr, value, target=False):
        if value and not isinstance(value, Instance):
            fullname = value.full_name()
        elif value and isinstance(value, Const):
            fullname = "%s" % value.get_value()
        elif isinstance(attr.parent, Instance):
            fullname = "%s%s" % (attr.parent_type.full_name(), not target and ".%s" % attr.name or "")
        else:
            fullname = "%s%s" % (attr.parent.full_name(), not target and ".%s" % attr.name or "")
        return fullname

    def _attributes_to_target_names(self, attributes):

        "Get the target names for the 'attributes'."

        output = []

        if attributes:
            for attr, value in attributes:
                fullname = self._attribute_value_to_name(attr, value, True)
                output.append(fullname)

        output.sort()
        return output

    def _attributes_to_attribute_names(self, attributes):

        "Get the output form of the 'attributes'."

        output = []

        if attributes:
            for attr, value in attributes:
                fullname = self._attribute_value_to_name(attr, value, False)
                output.append((fullname, value))

        output.sort()
        return output

    def _attribute_start(self, attrname, attributes):
        if attributes:
            self._span_start("attr")
            self._popup_start("attributes-popup")
            self._names_list_start("attributes", "attrnames")
            self._attribute_list(attributes)
            self._names_list_end()
            self._popup_end()

    def _attribute_list(self, attributes):

        # Mix links to attributes with labels indicating undetermined
        # attributes.

        last = None
        for fullname, value in attributes:
            if fullname != last:
                if last is not None:
                    self.stream.write("<br />")
                if value is not None and not isinstance(value, Instance):
                    self._object_name_ref(value.module, value, fullname, classes="attribute-name")
                else:
                    self.stream.write(self._text(fullname))
            last = fullname

    def _attribute_end(self, attributes):
        if attributes:
            self._span_end()

    def _get_possible_types(self, attrname):
        objtable = self.program.get_object_table()
        return objtable.any_possible_objects([attrname])

    def _get_attributes(self, possible_types, attrname):
        objtable = self.program.get_object_table()
        attributes = []
        for target_name in possible_types:
            target = objtable.get_object(target_name)
            try:
                attr = objtable.access(target_name, attrname)
            except TableError:
                continue
            if attr.is_static_attribute():
                for v in attr.get_values():
                    attributes.append((v, target, target_name))
            else:
                attributes.append((None, target, target_name))

        return attributes

# Summary classes.

class Summary(Writer):

    "Summarise classes and attributes in modules."

    def __init__(self, module, program):
        self.module = module
        self.program = program

    def to_stream(self, stream):

        "Write the summary to the given 'stream'."

        self.stream = stream
        self.stream.write(html_header % {
            "title" : "Module: %s" % self.module.full_name()
            })
        self._write_classes(self.module)
        self.stream.write(html_footer)

    def _write_classes(self, module):

        all_classes = {}

        for obj in self.module.all_objects:
            if isinstance(obj, Class):
                all_classes[obj.name] = obj

        if all_classes:

            all_class_names = all_classes.keys()
            all_class_names.sort()

            self.stream.write("<table cellspacing='5' cellpadding='5'>\n")
            self.stream.write("<thead>\n")
            self.stream.write("<tr>\n")
            self.stream.write("<th>Classes</th><th>Attributes</th>\n")
            self.stream.write("</tr>\n")
            self.stream.write("</thead>\n")

            for name in all_class_names:
                self._write_class(all_classes[name])

            self.stream.write("</table>\n")

    def _write_class(self, obj):

        # Write the class...

        self.stream.write("<tbody class='class'>\n")
        self.stream.write("<tr>\n")
        self.stream.write("<th class='summary-class' id='%s' rowspan='2'>" % self._attr(obj.full_name()))
        self._object_name_ref(self.module, obj, classes="class-name")
        self.stream.write("</th>\n")

        # ...and instance attribute names in order...

        attrs = obj.instance_attributes_as_list()

        if attrs:
            for attr in attrs:
                self.stream.write("<td class='summary-attr'>%s</td>\n" % self._text(attr.name))
        else:
            self.stream.write("<td class='summary-attr-absent'>None</td>\n")

        self.stream.write("</tr>\n")
        self.stream.write("<tr>\n")

        # ...and class attribute names in order.

        attrs = obj.attributes_as_list()

        if attrs:
            for attr in attrs:
                if attr.is_strict_constant():
                    value = attr.get_value()
                    if not isinstance(value, Const):
                        self.stream.write("<td class='summary-class-attr' id='%s'>" % self._attr(value.full_name()))
                        self._object_name_ref(self.module, value, attr.name, classes="summary-ref")
                        self.stream.write("</td>\n")
                    else:
                        self.stream.write("<td class='summary-class-attr'>%s</td>\n" % self._text(attr.name))
                else:
                    self.stream.write("<td class='summary-class-attr'>%s</td>\n" % self._text(attr.name))
        else:
            self.stream.write("<td class='summary-class-attr-absent'>None</td>\n")

        self.stream.write("</tr>\n")
        self.stream.write("</tbody>\n")

class Interfaces(Writer):

    "Summarise the interfaces used by reading the object table cache."

    def __init__(self, program):
        self.program = program

    def to_stream(self, stream):

        "Write the summary to the given 'stream'."

        self.stream = stream
        self.stream.write(html_header % {
            "title" : "Interfaces"
            })
        self._write_interfaces()
        self.stream.write(html_footer)

    def _write_interfaces(self):
        objtable = self.program.get_object_table()
        all_interfaces = objtable.all_cache.items()
        all_interfaces.sort()

        self.stream.write("<table cellspacing='5' cellpadding='5'>\n")
        self.stream.write("<thead>\n")
        self.stream.write("<tr>\n")
        self.stream.write("<th>Complete Interfaces</th>\n")
        self.stream.write("</tr>\n")
        self.stream.write("</thead>\n")
        self._write_interface_type(all_interfaces, "complete")
        self.stream.write("</table>\n")

    def _write_interface_type(self, interfaces, classes=""):
        self.stream.write("<tbody>\n")

        for names, objects in interfaces:
            if names:
                names = list(names)
                names.sort()
                self.stream.write("<tr>\n")
                self.stream.write("<td class='summary-interface %s'>%s</td>" % (classes, ", ".join(names)))
                self.stream.write("</tr>\n")

        self.stream.write("</tbody>\n")

# Source code classes.

class AnnotatedSource(ASTVisitor, Writer):

    "A module source code browser."

    def __init__(self, module, program):
        self.visitor = self
        self.module = module
        self.program = program
        self.units = []

    def get_unit(self):
        return self.units[-1]

    def to_stream(self, stream):

        "Write the annotated code to the given 'stream'."

        self.stream = stream
        self.stream.write(html_header % {
            "title" : "Module: %s" % self.module.full_name()
            })
        self.dispatch(self.module.astnode)
        self.stream.write(html_footer)

    def visitModule(self, node):
        self.units.append(node.unit)

        self._doc(node, "module")
        self.default(node)

        self.units.pop()

    # Statements.

    def visitAssert(self, node):
        self.stream.write("<div class='assert nowrap'>\n")
        self._keyword("assert")
        self.dispatch(node.test)
        if node.fail:
            self.stream.write(", ")
            self.dispatch(node.fail)
        self.stream.write("</div>\n")

    def visitAssign(self, node):
        self.stream.write("<div class='assign nowrap'>\n")
        for lvalue in node.nodes:
            self.dispatch(lvalue)
            self.stream.write(" = ")
        self.dispatch(node.expr)
        self.stream.write("</div>\n")

    def visitAugAssign(self, node):
        self.stream.write("<div class='augassign nowrap'>\n")
        self.dispatch(node.node)
        self._op(node.op, operator_functions[node.op], 1)
        self.dispatch(node.expr)
        self.stream.write("</div>\n")

    def visitBreak(self, node):
        self.stream.write("<div class='break nowrap'>\n")
        self._keyword("break")
        self.stream.write("</div>\n")

    def visitClass(self, node):
        if not used_by_unit(node):
            self._docstring('"Class %s not generated."' % node.name)
            return

        # Use inspected details where possible.

        cls = node.unit
        self.units.append(cls)

        bases = cls.bases
        self.stream.write("<div class='class nowrap' id='%s'>\n" % cls.full_name())

        # Write the declaration line.

        self.stream.write("<div>\n")
        self._keyword("class")
        self._object_name_def(self.module, cls, "class-name")

        # Suppress the "object" class appearing alone.

        if bases and not (len(bases) == 1 and bases[0].name == "object"):
            self.stream.write("(")
            first = True
            for base in bases:
                if not first:
                    self.stream.write(", ")

                self._object_name_ref(base.module, base)

                first = False
            self.stream.write(")")

        self.stream.write(":\n")
        self.stream.write("</div>\n")

        # Write the docstring and class body.

        self.stream.write("<div class='body nowrap'>\n")
        self._doc(node)

        # NOTE: Some streams may not support tell.

        x = self.stream.tell()

        self.default(node.code)

        # Check for no output.

        if x == self.stream.tell():
            self.visitPass(None)

        self.stream.write("</div>\n")
        self.stream.write("</div>\n")

        self.units.pop()

    def visitContinue(self, node):
        self.stream.write("<div class='continue nowrap'>\n")
        self._keyword("continue")
        self.stream.write("</div>\n")

    def visitDiscard(self, node):
        self.stream.write("<div class='discard nowrap'>\n")
        self.default(node)
        self.stream.write("</div>\n")

    def visitExec(self, node):
        self.stream.write("<div class='exec nowrap'>\n")
        self._keyword("exec")
        self.dispatch(node.expr)
        if node.locals:
            self._keyword("in", 1)
            self.dispatch(node.locals)
        if node.globals:
            self.stream.write(", ")
            self.dispatch(node.globals)
        self.stream.write("</div>\n")

    def visitFor(self, node):
        self.stream.write("<div class='if nowrap'>\n")
        self.stream.write("<div>\n")
        self._keyword("for")
        self.dispatch(node.assign)
        self._keyword("in", 1)
        self.dispatch(node.list)
        self.stream.write(":\n")
        self.stream.write("</div>\n")
        self.stream.write("<div class='body nowrap'>\n")
        self.dispatch(node.body)
        self.stream.write("</div>\n")
        if node.else_ is not None:
            self.stream.write("<div>\n")
            self._keyword("else", trailing=0)
            self.stream.write(":\n")
            self.stream.write("</div>\n")
            self.stream.write("<div class='body nowrap'>\n")
            self.dispatch(node.else_)
            self.stream.write("</div>\n")
        self.stream.write("</div>\n")

    def visitFrom(self, node):
        self.stream.write("<div class='from nowrap'>\n")
        self._keyword("from")

        # Support relative imports of names and whole modules.

        modname, names = get_module_name(node, self.module)
        if modname:
            self._module_link(modname, "%s%s" % ("." * node.level, modname))
        else:
            self._text("." * node.level)

        self._keyword("import", 1)
        first = True

        # Integrate the provided name details with any calculated details
        # resulting from whole module imports.

        for (name, alias), (full_name, alias) in zip(node.names, names):
            if not first:
                self.stream.write(", ")

            if modname:
                self._name(name)
            else:
                self._module_link(full_name, name)

            if alias:
                self._keyword("as", 1)
                self._name(alias)
            first = False

        self.stream.write("</div>\n")

    def visitFunction(self, node):
        if not used_by_unit(node):
            self._docstring('"Function %s not generated."' % node.name)
            return

        fn = node.unit
        self.units.append(fn)

        self.stream.write("<div class='function nowrap' id='%s'>\n" % fn.full_name())

        # Write the declaration line.

        self.stream.write("<div>\n")
        self._keyword("def")
        self._object_name_def(self.module, fn, "function-name")

        self.stream.write("(")
        self._parameters(fn, node)
        self.stream.write(")")
        self.stream.write(":\n")
        self.stream.write("</div>\n")

        self.stream.write("<div class='body nowrap'>\n")
        self._doc(node)
        self.dispatch(node.code)
        self.stream.write("</div>\n")
        self.stream.write("</div>\n")

        self.units.pop()

    def visitGlobal(self, node):
        self.stream.write("<div class='global nowrap'>\n")
        self._keyword("global")
        first = True
        for name in node.names:
            if not first:
                self.stream.write(", ")
            self.stream.write(name)
            first = False
        self.stream.write("</div>\n")

    def visitIf(self, node):
        self.stream.write("<div class='if nowrap'>\n")
        first = True
        for compare, stmt in node.tests:
            self.stream.write("<div>\n")
            if first:
                self._keyword("if")
            else:
                self._keyword("elif")
            self.dispatch(compare)
            self.stream.write(":\n")
            self.stream.write("</div>\n")
            self.stream.write("<div class='body nowrap'>\n")
            self.dispatch(stmt)
            self.stream.write("</div>\n")
            first = False
        if node.else_ is not None:
            self.stream.write("<div>\n")
            self._keyword("else", trailing=0)
            self.stream.write(":\n")
            self.stream.write("</div>\n")
            self.stream.write("<div class='body nowrap'>\n")
            self.dispatch(node.else_)
            self.stream.write("</div>\n")
        self.stream.write("</div>\n")

    def visitImport(self, node):
        self.stream.write("<div class='import nowrap'>\n")
        self._keyword("import")
        first = True
        for name, alias in node.names:
            if not first:
                self.stream.write(",\n")
            self._module_link(name, name)
            if alias:
                self._keyword("as", 1)
                self._name(alias)
            first = False
        self.stream.write("</div>\n")

    def visitPass(self, node):
        self.stream.write("<div class='pass nowrap'>\n")
        self._keyword("pass")
        self.stream.write("</div>\n")

    def visitPrint(self, node):
        self.stream.write("<div class='print nowrap'>\n")
        self._keyword("print")
        if node.dest is not None:
            self.stream.write(">>\n")
            self.dispatch(node.dest)
            self.stream.write(",\n")
        for n in node.nodes:
            self.dispatch(n)
            self.stream.write(",\n")
        self.stream.write("</div>\n")

    def visitPrintnl(self, node):
        self.stream.write("<div class='printnl nowrap'>\n")
        self._keyword("print")
        if node.dest is not None:
            self.stream.write(">>\n")
            self.dispatch(node.dest)
            first = False
        else:
            first = True
        for n in node.nodes:
            if not first:
                self.stream.write(",\n")
            self.dispatch(n)
            first = False
        self.stream.write("</div>\n")

    def visitRaise(self, node):
        self.stream.write("<div class='raise nowrap'>\n")
        self._keyword("raise")
        if node.expr1 is not None:
            self.dispatch(node.expr1)
        if node.expr2 is not None:
            self.stream.write(",\n")
            self.dispatch(node.expr2)
        if node.expr3 is not None:
            self.stream.write(",\n")
            self.dispatch(node.expr3)
        self.stream.write("</div>\n")

    def visitReturn(self, node):
        self.stream.write("<div class='return nowrap'>\n")
        self._keyword("return")
        self.dispatch(node.value)
        self.stream.write("</div>\n")

    def visitStmt(self, node):
        self.stream.write("<div class='stmt nowrap'>\n")
        self.default(node)
        self.stream.write("</div>\n")

    def visitTryExcept(self, node):
        self.stream.write("<div class='tryexcept nowrap'>\n")
        self.stream.write("<div>\n")
        self._keyword("try", trailing=0)
        self.stream.write(":\n")
        self.stream.write("</div>\n")
        self.stream.write("<div class='body nowrap'>\n")
        self.dispatch(node.body)
        self.stream.write("</div>\n")
        for spec, assign, statement in node.handlers:
            self.stream.write("<div>\n")
            self._keyword("except")
            if spec is not None:
                self.dispatch(spec)
            if assign is not None:
                self.stream.write(",\n")
                self.dispatch(assign)
            self.stream.write(":\n")
            self.stream.write("</div>\n")
            self.stream.write("<div class='body nowrap'>\n")
            self.dispatch(statement)
            self.stream.write("</div>\n")
        if node.else_ is not None:
            self.stream.write("<div>\n")
            self._keyword("else", trailing=0)
            self.stream.write(":\n")
            self.stream.write("</div>\n")
            self.stream.write("<div class='body nowrap'>\n")
            self.dispatch(node.else_)
            self.stream.write("</div>\n")
        self.stream.write("</div>\n")

    def visitTryFinally(self, node):
        self.stream.write("<div class='tryfinally nowrap'>\n")
        self.stream.write("<div>\n")
        self._keyword("try", trailing=0)
        self.stream.write(":\n")
        self.stream.write("</div>\n")
        self.stream.write("<div class='body nowrap'>\n")
        self.dispatch(node.body)
        self.stream.write("</div>\n")
        self.stream.write("<div>\n")
        self._keyword("finally", trailing=0)
        self.stream.write(":\n")
        self.stream.write("</div>\n")
        self.stream.write("<div class='body nowrap'>\n")
        self.dispatch(node.final)
        self.stream.write("</div>\n")
        self.stream.write("</div>\n")

    def visitWhile(self, node):
        self.stream.write("<div class='while nowrap'>\n")
        self.stream.write("<div>\n")
        self._keyword("while")
        self.dispatch(node.test)
        self.stream.write(":\n")
        self.stream.write("</div>\n")
        self.stream.write("<div class='body nowrap'>\n")
        self.dispatch(node.body)
        self.stream.write("</div>\n")
        if node.else_ is not None:
            self.stream.write("<div>\n")
            self._keyword("else", trailing=0)
            self.stream.write(":\n")
            self.stream.write("</div>\n")
            self.stream.write("<div class='body nowrap'>\n")
            self.dispatch(node.else_)
            self.stream.write("</div>\n")
        self.stream.write("</div>\n")

    def visitYield(self, node):
        self.stream.write("<div class='yield nowrap'>\n")
        self._keyword("yield")
        self.dispatch(node.value)
        self.stream.write("</div>\n")

    # Expression-related helper methods.

    def _visitBitBinary(self, node, symbol):
        name = operator_functions[node.__class__.__name__]
        self._span_start(name)
        first = True
        for node in node.nodes:
            if not first:
                self._op(symbol, name, 1)
            self.dispatch(node)
            first = False
        self._span_end()

    def _visitBinary(self, node, symbol):
        name = operator_functions[node.__class__.__name__]
        self._span_start(name)
        self.dispatch(node.left)
        self._op(symbol, name, 1)
        self.dispatch(node.right)
        self._span_end()

    def _visitUnary(self, node, symbol):
        name = operator_functions[node.__class__.__name__]
        self._span_start(name)
        self._op(symbol, name, trailing=0)
        self.dispatch(node.expr)
        self._span_end()

    # Expressions.

    def visitAdd(self, node):
        self._visitBinary(node, "+")

    def visitAnd(self, node):
        self._span_start("and")
        first = True
        for n in node.nodes:
            if not first:
                self._keyword("and", 1)
            self.dispatch(n)
            first = False
        self._span_end()

    def _visitAttr(self, node, label):
        self.record_unknown_targets(node)

        attributes = node._value_deduced and [self.get_attribute_and_value(node._value_deduced)] or \
                     node._attr_deduced and [self.get_attribute_and_value(node._attr_deduced)] or \
                     node._attrs_deduced or \
                     map(self.get_attribute_and_value, node._attrs_deduced_from_specific_usage or [])

        possible_types = self._attributes_to_target_names(attributes)

        wraps_getattr = self._has_descendant(node.expr, compiler.ast.Getattr)

        if not wraps_getattr:
            self._span_start(label)
            self._accessor_start(possible_types)
        self.dispatch(node.expr)
        if not wraps_getattr:
            self._accessor_end(possible_types)

        self.stream.write(".")
        self._attribute_start(node.attrname, self._attributes_to_attribute_names(attributes))
        self._span(node.attrname, "attrname" + (not possible_types and " no-targets" or ""))
        self._attribute_end(attributes)

        if not wraps_getattr:
            self._span_end()

    def visitAssAttr(self, node):
        if node.flags == "OP_DELETE":
            self._keyword("del")
        self._visitAttr(node, "assattr")

    def visitAssList(self, node):
        self._span_start("list")
        self.stream.write("[")
        self._sequence(node)
        self.stream.write("]")
        self._span_end()

    def visitAssName(self, node):
        self._assname(node.name, node)

    def visitAssTuple(self, node):
        self._span_start("tuple")
        self.stream.write("(")
        self._sequence(node)
        self.stream.write(")")
        self._span_end()

    def visitBackquote(self, node):
        self._span_start("backquote")
        self.stream.write("`")
        self.dispatch(node.expr)
        self.stream.write("`")
        self._span_end()

    def visitBitand(self, node):
        self._visitBitBinary(node, "&")

    def visitBitor(self, node):
        self._visitBitBinary(node, "|")

    def visitBitxor(self, node):
        self._visitBitBinary(node, "^")

    def visitCallFunc(self, node):
        self._span_start("callfunc")
        self.dispatch(node.node)
        self._span_start("call")
        self.stream.write("(")
        first = True
        for arg in node.args:
            if not first:
                self.stream.write(", ")
            self.dispatch(arg)
            first = False
        if node.star_args is not None:
            if not first:
                self.stream.write(", *")
            self.dispatch(node.star_args)
            first = False
        if node.dstar_args is not None:
            if not first:
                self.stream.write(", **")
            self.dispatch(node.dstar_args)
            first = False
        self.stream.write(")")
        self._span_end()
        self._span_end()

    def visitCompare(self, node):
        self._span_start("compare")
        self.dispatch(node.expr)
        for op_name, expr in node.ops:
            self._op(op_name, operator_functions.get(op_name), 1)
            self.dispatch(expr)
        self._span_end()

    def visitConst(self, node):
        if isinstance(node.value, (str, unicode)):
            self._span_start("str")
        self.stream.write(self._text(repr(node.value)))
        if isinstance(node.value, (str, unicode)):
            self._span_end()

    def visitDict(self, node):
        self._span_start("dict")
        self.stream.write("{")
        self._mapping(node)
        self.stream.write("}")
        self._span_end()

    def visitDiv(self, node):
        self._visitBinary(node, "/")

    def visitFloorDiv(self, node):
        self._visitBinary(node, "//")

    def visitGetattr(self, node):
        self._visitAttr(node, "getattr")

    def visitGenExpr(self, node):
        self._span_start("genexpr")
        self.stream.write("(")
        self.dispatch(node.code)
        self.stream.write(")")
        self._span_end()

    def visitGenExprFor(self, node):
        self._span_start("genexprfor")
        self._keyword("for", 1)
        self._span_start("item")
        self.dispatch(node.assign)
        self._span_end()
        self._keyword("in", 1)
        self._span_start("collection")
        self.dispatch(node.iter)
        self._span_end()
        for if_ in node.ifs:
            self.dispatch(if_)
        self._span_end()

    def visitGenExprIf(self, node):
        self._span_start("genexprif")
        self._span_start("conditional")
        self._keyword("if", 1)
        self.dispatch(node.test)
        self._span_end()
        self._span_end()

    def visitGenExprInner(self, node):
        self._span_start("genexprinner")
        self.dispatch(node.expr)
        for qual in node.quals:
            self.dispatch(qual)
        self._span_end()

    def visitIfExp(self, node):
        self._span_start("ifexp")
        self.dispatch(node.then)
        self._keyword("if", 1)
        self.dispatch(node.test)
        self._keyword("else", 1)
        self.dispatch(node.else_)
        self._span_end()

    def visitInvert(self, node):
        self._visitUnary(node, "~")

    def visitKeyword(self, node):
        self._span_start("keyword-arg")
        self.stream.write(node.name)
        self.stream.write("=")
        self.dispatch(node.expr)
        self._span_end()

    def visitLambda(self, node):
        fn = node.unit
        self.units.append(fn)

        self._span_start("lambda")
        self._keyword("lambda")
        self._parameters(fn, node)
        self.stream.write(": ")
        self._span_start("code")
        self.dispatch(node.code)
        self._span_end()
        self._span_end()

        self.units.pop()

    def visitLeftShift(self, node):
        self._visitBinary(node, "<<")

    visitList = visitAssList

    def visitListComp(self, node):
        self._span_start("listcomp")
        self.stream.write("[")
        self.dispatch(node.expr)
        for qual in node.quals:
            self.dispatch(qual)
        self.stream.write("]")
        self._span_end()

    def visitListCompFor(self, node):
        self._span_start("listcompfor")
        self._keyword("for", 1)
        self._span_start("item")
        self.dispatch(node.assign)
        self._span_end()
        self._keyword("in", 1)
        self._span_start("collection")
        self.dispatch(node.list)
        self._span_end()
        for if_ in node.ifs:
            self.dispatch(if_)
        self._span_end()

    def visitListCompIf(self, node):
        self._span_start("listcompif")
        self._span_start("conditional")
        self._keyword("if", 1)
        self.dispatch(node.test)
        self._span_end()
        self._span_end()

    def visitMod(self, node):
        self._visitBinary(node, "%")

    def visitMul(self, node):
        self._visitBinary(node, "*")

    def visitName(self, node):
        if node._scope:
            scope = node._scope
            self._name_start()
            self.stream.write(node.name)
            self._popup_start()
            self._scope(node._scope, node._attr)
            self._popup_end()
            self._name_end()
        else:
            self._span(node.name)

    def visitNot(self, node):
        self._span_start("not")
        self._keyword("not")
        self.dispatch(node.expr)
        self._span_end()

    def visitOr(self, node):
        self._span_start("or")
        first = True
        for n in node.nodes:
            if not first:
                self._keyword("or", 1)
            self.dispatch(n)
            first = False
        self._span_end()

    def visitPower(self, node):
        self._visitBinary(node, "**")

    def visitRightShift(self, node):
        self._visitBinary(node, ">>")

    def visitSlice(self, node):
        self._span_start("slice")
        self.dispatch(node.expr)
        self.stream.write("[")
        if node.lower:
            self.dispatch(node.lower)
        self.stream.write(":")
        if node.upper:
            self.dispatch(node.upper)
        # NOTE: Step?
        self.stream.write("]")
        self._span_end()

    def visitSliceobj(self, node):
        self._span_start("sliceobj")
        first = True
        for n in node.nodes:
            if not first:
                self.stream.write(":")
            self.dispatch(n)
        self._span_end()

    def visitSub(self, node):
        self._visitBinary(node, "-")

    def visitSubscript(self, node):
        self._span_start("subscript")
        self.dispatch(node.expr)
        self.stream.write("[")
        first = True
        for sub in node.subs:
            if not first:
                self.stream.write(", ")
            self.dispatch(sub)
            first = False
        self.stream.write("]")
        self._span_end()

    visitTuple = visitAssTuple

    def visitUnaryAdd(self, node):
        self._visitUnary(node, "+")

    def visitUnarySub(self, node):
        self._visitUnary(node, "-")

    # Output preparation methods.

    def _sequence(self, node):
        first = True
        for n in node.nodes:
            if not first:
                self.stream.write(", ")
            self.dispatch(n)
            first = False

    def _mapping(self, node):
        first = True
        for k, v in node.items:
            if not first:
                self.stream.write(", ")
            self.dispatch(k)
            self.stream.write(" : ")
            self.dispatch(v)
            first = False

    def _parameters(self, fn, node):
        nparams = len(fn.positional_names)
        ndefaults = len(fn.defaults)
        first_with_default = nparams - ndefaults

        first = True
        for n, param in enumerate(fn.positional_names):
            if not first:
                self.stream.write(", ")

            # Handle tuple parameters.

            if isinstance(param, tuple):
                self._tuple_parameter(param, node)
            else:
                self._assname(param, node)

            n_default = n - first_with_default
            if n_default >= 0:
                self._default(fn.defaults[n_default])
            first = False

        if fn.has_star:
            if not first:
                self.stream.write(", *")
            self._name(fn.star_name)

        if fn.has_dstar:
            if not first:
                self.stream.write(", **")
            self._name(fn.dstar_name)

    def _tuple_parameter(self, parameters, node):
        self.stream.write("(")

        first = True
        for param in parameters:
            if not first:
                self.stream.write(", ")

            # Handle tuples.

            if isinstance(param, tuple):
                self._tuple_parameter(param, node)
            else:
                self._assname(param, node)

            first = False

        self.stream.write(")")

    def _default(self, default):
        self.stream.write("=")
        self.dispatch(default)

    # Statistics gathering methods.

    def record_unknown_targets(self, node):
        if not node._attrs_deduced:
            self.program.unknown_target_nodes.append((self.units[-1], node))
        elif not node._attrs_deduced_from_specific_usage:
            self.program.independent_target_nodes.append((self.units[-1], node))

    # Utility methods.

    def _has_descendant(self, node, nodetype):
        if isinstance(node, nodetype):
            return True
        else:
            for n in node.getChildNodes():
                if self._has_descendant(n, nodetype):
                    return True
            return False

# Convenience functions.

def summarise(module, program, filename):
    stream = open(filename, "wb")
    try:
        summary = Summary(module, program)
        summary.to_stream(stream)
    finally:
        stream.close()

def annotate(module, program, filename):
    stream = open(filename, "wb")
    try:
        source = AnnotatedSource(module, program)
        source.to_stream(stream)
    finally:
        stream.close()

def interfaces(program, filename):
    stream = open(filename, "wb")
    try:
        source = Interfaces(program)
        source.to_stream(stream)
    finally:
        stream.close()

def report(program, directory):
    if not exists(directory):
        os.mkdir(directory)

    for module in program.get_importer().get_modules():
        annotate(module, program, join(directory, "%s%sxhtml" % (module.full_name(), extsep)))
        summarise(module, program, join(directory, "%s-summary%sxhtml" % (module.full_name(), extsep)))

    interfaces(program, join(directory, "-interfaces%sxhtml" % extsep))

# vim: tabstop=4 expandtab shiftwidth=4
